import struct
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING
from ..rpc.request import RpcRequest
from ..rpc.response import RpcResponse
from ..uuid import uuid_to_bin
from .orpcthis import ORPCTHIS


if TYPE_CHECKING:
    from ..protocol import Protocol


class NdrInterface(ABC):

    @abstractmethod
    def get_ipid(self) -> int:
        ...

    async def rem_release(self, proto: 'Protocol'):
        c_public_refs = 1
        c_private_refs = 0

        orpcthis = ORPCTHIS.get_data(flags=0)
        fixed = b'\x01\x00\xce\xce\x01\x00\x00\x00'
        element = uuid_to_bin(self.get_ipid()) + struct.pack(
            '<LL',
            c_public_refs,
            c_private_refs)

        pdu_data = orpcthis + fixed + element

        iremunknown = \
            proto._interface.scm_reply_info_data.ipid_rem_unknown

        request = RpcRequest(op_num=5, uuid_str=iremunknown)
        request.set_pdu_data(pdu_data)

        request_pkg = request.sign_data(proto)

        # rem release shouldn't take long so we can use the default timeout
        rpc_response: RpcResponse = \
            await proto.get_dcom_response(request_pkg, RpcResponse.SIZE)

        message = rpc_response.get_message(proto)



import functools
import os
import hashlib
import hmac
import math
import struct
from functools import reduce
from pyasn1.codec.der import decoder, encoder
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from Crypto.Cipher import AES
from Crypto.Hash import HMAC, SHA1, MD5
from .rc4 import RC4
from .gss import gss_wrap_rc4, gss_unwrap_rc4
from ..exceptions import KerberosErr


def _nfold(ba, nbytes):
    def rotate_right(ba, nbits):
        ba = bytearray(ba)
        n = len(ba)
        nbits %= (n * 8)
        # Rotate bytes
        for _ in range(nbits // 8):
            ba = bytearray([ba[-1]]) + ba[:-1]
        # Rotate remaining bits
        nbits %= 8
        if nbits > 0:
            last_bit = 0
            for i in range(len(ba)):
                new_last_bit = ba[i] & ((1 << nbits) - 1)
                ba[i] = (ba[i] >> nbits) | (last_bit << (8 - nbits))
                last_bit = new_last_bit
            ba[0] |= (last_bit << (8 - nbits))
        return ba

    def add_ones_complement(ba1, ba2):
        n = len(ba1)
        res = [0] * n
        carry = 0
        for i in range(n-1, -1, -1):
            s = ba1[i] + ba2[i] + carry
            res[i] = s & 0xff
            carry = s >> 8

        pos = n - 1
        while carry and pos >= 0:
            s = res[pos] + carry
            res[pos] = s & 0xff
            carry = s >> 8
            pos -= 1
        return bytearray(res)

    slen = len(ba)

    def lcm(a, b):
        return abs(a*b) // math.gcd(a, b)

    lcm_val = lcm(slen, nbytes)
    big_str = bytearray()
    curr = bytearray(ba)
    for i in range(lcm_val // slen):
        big_str += curr
        curr = rotate_right(curr, 13)

    parts = [big_str[i:i+nbytes] for i in range(0, len(big_str), nbytes)]
    return bytes(reduce(add_ones_complement, parts))


def derive_key(base_key, usage_int, payload_byte=None):
    constant = struct.pack('>I', usage_int)
    if payload_byte is not None:
        constant += bytes([payload_byte])

    nfolded = _nfold(constant, 16)
    cipher = Cipher(algorithms.AES(base_key), modes.ECB())
    encryptor1 = cipher.encryptor()
    b1 = encryptor1.update(nfolded) + encryptor1.finalize()
    encryptor2 = cipher.encryptor()
    b2 = encryptor2.update(b1) + encryptor2.finalize()

    return b1 + b2


def aes_cts_encrypt(key: bytes, plain: bytes) -> bytes:
    n = len(plain)  # 43 bytes
    iv = b'\x00' * 16
    pad_len = 16 - (n % 16)
    padded_plain = plain + b'\x00' * pad_len

    # CBC encryptie (correct IV)
    cipher = Cipher(algorithms.AES(key), modes.CBC(iv))
    encryptor = cipher.encryptor()
    full_cipher = encryptor.update(padded_plain) + encryptor.finalize()

    c1 = full_cipher[0:16]
    c2 = full_cipher[16:32]
    c3 = full_cipher[32:48]

    return c1 + c3 + c2[:n % 16]


def decrypt_kerberos_aes_cts(key: bytes, usage: int, cipher: bytes) -> bytes:
    ciphertext = cipher[:-12]
    expected_hmac = cipher[-12:]

    ki = derive_key(key, usage, 0x55)
    ke = derive_key(key, usage, 0xAA)

    n = len(ciphertext)
    block_size = 16

    aes = AES.new(ke, AES.MODE_ECB)

    if n % block_size == 0:
        aes_cbc = AES.new(ke, AES.MODE_CBC, b'\x00' * block_size)
        plaintext_with_confounder = aes_cbc.decrypt(ciphertext)
    else:
        m = n // block_size
        last_blocks_len = n % block_size

        iv = b'\x00' * block_size
        plaintext_with_confounder = b''

        for i in range(m - 1):
            block = ciphertext[i*block_size: (i+1)*block_size]
            dec = aes.decrypt(block)
            plaintext_with_confounder += \
                bytes(a ^ b for a, b in zip(dec, iv))
            iv = block

        cn_minus_1 = ciphertext[(m-1)*block_size: m*block_size]
        cn = ciphertext[m*block_size:]

        dec_intermediate = aes.decrypt(cn_minus_1)

        pn = bytes(
            a ^ b for a, b in zip(dec_intermediate[:last_blocks_len], cn))

        stolen_bytes = dec_intermediate[last_blocks_len:]
        full_block_to_decrypt = cn + stolen_bytes

        dec_pn_minus_1 = aes.decrypt(full_block_to_decrypt)
        pn_minus_1 = bytes(a ^ b for a, b in zip(dec_pn_minus_1, iv))

        plaintext_with_confounder += pn_minus_1 + pn

    h = HMAC.new(ki, plaintext_with_confounder, SHA1)
    if h.digest()[:12] != expected_hmac:
        raise ValueError("Integrity check failed: HMAC mismatch")

    return plaintext_with_confounder


def encrypt_kerberos_aes_cts(session_key: bytes,
                             usage: int,
                             plain_text: bytes):
    ki = derive_key(session_key, usage, 0x55)
    ke = derive_key(session_key, usage, 0xAA)

    confounder = os.urandom(16)
    basic_plaintext = confounder + plain_text

    h = HMAC.new(ki, basic_plaintext, SHA1)
    checksum = h.digest()[:12]

    aes = AES.new(ke, AES.MODE_CBC, b'\x00' * 16)
    n = len(basic_plaintext)

    if n % 16 == 0:
        final_ctext = aes.encrypt(basic_plaintext)
    else:
        pad_len = 16 - (n % 16)
        padded_data = basic_plaintext + b'\x00' * pad_len
        ctext = aes.encrypt(padded_data)

        last_full_block = ctext[-32:-16]
        truncated_block = ctext[-16:]

        final_ctext = (
            ctext[:-32] +
            truncated_block +
            last_full_block[:16-pad_len]
        )

    return final_ctext + checksum


def encrypt_kerberos_rc4(session_key, usage, plaintext):
    confounder = os.urandom(8)
    data_to_encrypt = confounder + plaintext

    usage_bytes = struct.pack('<I', usage)
    k1 = HMAC.new(session_key, usage_bytes, MD5).digest()

    checksum = HMAC.new(k1, data_to_encrypt, MD5).digest()
    k3 = HMAC.new(k1, checksum, MD5).digest()

    cipher = RC4(k3)
    encrypted_data = cipher.encrypt(data_to_encrypt)

    return checksum + encrypted_data


def decrypt_kerberos_rc4(key, usage, data):
    checksum = data[:16]
    encrypted_data = data[16:]

    k1 = hmac.new(key, struct.pack('<I', usage), hashlib.md5).digest()
    k3 = hmac.new(k1, checksum, hashlib.md5).digest()

    cipher = RC4(k3)
    decrypted = cipher.decrypt(encrypted_data)

    return decrypted[8:]  # Skip 8 bytes confounder


def read_session_key(data: bytes) -> bytes:
    payload = data[16:]
    as_rep_part, _ = decoder.decode(payload)
    key_container_bytes = encoder.encode(as_rep_part[0])
    inner_key_seq, _ = decoder.decode(key_container_bytes)

    key_type_bytes = encoder.encode(inner_key_seq[0])
    key_value_bytes = encoder.encode(inner_key_seq[1])

    key_type_obj, _ = decoder.decode(key_type_bytes)
    key_value_obj, _ = decoder.decode(key_value_bytes)

    session_key = bytes(key_value_obj.asOctets())
    _key_type = int(key_type_obj)

    return session_key


def seal_func_kerberos(session_key: bytes):
    def _kerberos_sealer(
            flags: int,
            seq_num: int,
            message_to_sign: bytes,
            message_to_encrypt: bytes,
            session_key: bytes) -> tuple[bytes, bytes]:
        return gss_wrap_rc4(session_key,
                            message_to_encrypt,
                            seq_num)
    return functools.partial(_kerberos_sealer, session_key=session_key)


def gss_unwrap_kerberos(session_key: bytes):
    def _kerberos_unwrap(
            cipher_text: bytes,
            auth_data: bytes,
            session_key: bytes) -> tuple[bytes, bytes]:
        return gss_unwrap_rc4(session_key,
                              cipher_text,
                              auth_data)
    return functools.partial(_kerberos_unwrap, session_key=session_key)


def sign_func_kerberos(session_key: bytes):
    def _kerberos_signer(
            flags: int,
            seq_num: int,
            message_to_sign: bytes,
            session_key: bytes) -> tuple[bytes, bytes]:
        return gss_wrap_rc4(session_key,
                            message_to_sign,
                            seq_num)
    return functools.partial(_kerberos_signer, session_key=session_key)


KDC_ERR_PREAUTH_REQUIRED = 25

KRB_ERRORS = {
    0:  "KDC_ERR_NONE (No error)",
    1:  "KDC_ERR_NAME_EXP (Client's entry in database has expired)",
    2:  "KDC_ERR_SERVICE_EXP (Server's entry in database has expired)",
    3:  "KDC_ERR_BAD_PVNO (Requested protocol version number not supported)",
    4:  "KDC_ERR_C_OLD_MAST_KVNO (Client's key encrypted in old master key)",
    5:  "KDC_ERR_S_OLD_MAST_KVNO (Server's key encrypted in old master key)",
    6:  "KDC_ERR_C_PRINCIPAL_UNKNOWN (Username not found in Kerberos db)",
    7:  "KDC_ERR_S_PRINCIPAL_UNKNOWN (Server not found; Must use FQDN)",
    8:  "KDC_ERR_PRINCIPAL_NOT_UNIQUE (Multiple principal entries in db)",
    9:  "KDC_ERR_NULL_KEY (The client or server has a null key)",
    10: "KDC_ERR_CANNOT_POSTDATE (Ticket not eligible for postdating)",
    11: "KDC_ERR_NEVER_VALID (Requested starttime is later than end time)",
    12: "KDC_ERR_POLICY (KDC policy rejects request)",
    13: "KDC_ERR_BADOPTION (KDC cannot accommodate requested option)",
    14: "KDC_ERR_ETYPE_NOSUPP (KDC has no support for encryption type)",
    15: "KDC_ERR_SUMTYPE_NOSUPP (KDC has no support for checksum type)",
    16: "KDC_ERR_PADATA_TYPE_NOSUPP (KDC has no support for padata type)",
    17: "KDC_ERR_TRTYPE_NOSUPP (KDC has no support for transited type)",
    18: "KDC_ERR_CLIENT_REVOKED (Clients credentials have been revoked)",
    19: "KDC_ERR_SERVICE_REVOKED (Credentials for server have been revoked)",
    20: "KDC_ERR_TGT_REVOKED (TGT has been revoked)",
    21: "KDC_ERR_CLIENT_NOTYET (Client not yet valid; try again later)",
    22: "KDC_ERR_SERVICE_NOTYET (Server not yet valid; try again later)",
    23: "KDC_ERR_KEY_EXPIRED (Password has expired; change password to reset)",
    24: "KDC_ERR_PREAUTH_FAILED (Pre-auth invalid; check password)",
    25: "KDC_ERR_PREAUTH_REQUIRED (Additional pre-authentication required)",
    26: "KDC_ERR_SERVER_NOMATCH (Requested server and ticket don't match)",
    27: "KDC_ERR_MUST_USE_USER2USER (Server principal valid for usr2usr only)",
    28: "KDC_ERR_PATH_NOT_ACCEPTED (KDC Policy rejects transited path)",
    29: "KDC_ERR_SVC_UNAVAILABLE (A service is not available)",
    31: "KRB_AP_ERR_BAD_INTEGRITY (Integrity check on decrypted field failed)",
    32: "KRB_AP_ERR_TKT_EXPIRED (Ticket expired)",
    33: "KRB_AP_ERR_TKT_NYV (Ticket not yet valid)",
    34: "KRB_AP_ERR_REPEAT (Request is a replay)",
    35: "KRB_AP_ERR_NOT_US (The ticket isn't for us)",
    36: "KRB_AP_ERR_BADMATCH (Ticket/Auth mismatch; check user@domain (FQDN))",
    37: "KRB_AP_ERR_SKEW (Clock skew too great)",
    38: "KRB_AP_ERR_BADADDR (Incorrect net address)",
    39: "KRB_AP_ERR_BADVERSION (Protocol version mismatch)",
    40: "KRB_AP_ERR_MSG_TYPE (Invalid msg type)",
    41: "KRB_AP_ERR_MODIFIED (Message stream modified)",
    42: "KRB_AP_ERR_BADORDER (Message out of order)",
    44: "KRB_AP_ERR_BADKEYVER (Specified version of key is not available)",
    45: "KRB_AP_ERR_NOKEY (Service key not available)",
    46: "KRB_AP_ERR_MUT_FAIL (Mutual authentication failed)",
    47: "KRB_AP_ERR_BADDIRECTION (Incorrect message direction)",
    48: "KRB_AP_ERR_METHOD (Alternative authentication method required)",
    49: "KRB_AP_ERR_BADSEQ (Incorrect sequence number in message)",
    50: "KRB_AP_ERR_INAPP_CKSUM (Inappropriate type of checksum in message)",
    51: "KRB_AP_PATH_NOT_ACCEPTED (Policy rejects transited path)",
    52: "KRB_ERR_RESPONSE_TOO_BIG (Response too big for UDP; retry with TCP)",
    60: "KRB_ERR_GENERIC (Generic error; description in e-text)",
    61: "KRB_ERR_FIELD_TOOLONG (Field is too long for this implementation)",
    62: "KDC_ERROR_CLIENT_NOT_TRUSTED (Reserved for PKINIT)",
    63: "KDC_ERROR_KDC_NOT_TRUSTED (Reserved for PKINIT)",
    64: "KDC_ERROR_INVALID_SIG (Reserved for PKINIT)",
    65: "KDC_ERR_KEY_TOO_WEAK (Reserved for PKINIT)",
    66: "KDC_ERR_CERTIFICATE_MISMATCH (Reserved for PKINIT)",
    67: "KRB_AP_ERR_NO_TGT (No TGT available to validate USER-TO-USER)",
    68: "KDC_ERR_WRONG_REALM (Reserved for future use)",
    69: "KRB_AP_ERR_USER_TO_USER_REQUIRED (Ticket must be for USER-TO-USER)",
    70: "KDC_ERR_CANT_VERIFY_CERTIFICATE (Reserved for PKINIT)",
    71: "KDC_ERR_INVALID_CERTIFICATE (Reserved for PKINIT)",
    72: "KDC_ERR_REVOKED_CERTIFICATE (Reserved for PKINIT)",
    73: "KDC_ERR_REVOCATION_STATUS_UNKNOWN (Reserved for PKINIT)",
    74: "KDC_ERR_REVOCATION_STATUS_UNAVAILABLE (Reserved for PKINIT)",
    75: "KDC_ERR_CLIENT_NAME_MISMATCH (Reserved for PKINIT)",
    76: "KDC_ERR_KDC_NAME_MISMATCH (Reserved for PKINIT)",
    77: (
        "KDC_ERR_INCONSISTENT_KEY_PURPOSE "
        "(Certificate cannot be used for PKINIT client authentication)"),
    78: (
        "KDC_ERR_DIGEST_IN_CERT_NOT_ACCEPTED "
        "(Digest algorithm for the public key is not acceptable)"),
    79: "KDC_ERR_PA_CHECKSUM_MUST_BE_INCLUDED (The paChksum field is missing)",
    80: (
        "KDC_ERR_DIGEST_IN_SIGNED_DATA_NOT_ACCEPTED "
        "(Digest algorithm used by id-pkinit-authData is not acceptable)"),
    81: (
        "KDC_ERR_PUBLIC_KEY_ENCRYPTION_NOT_SUPPORTED "
        "(The KDC does not support public key encryption delivery)"),
    90: "KDC_ERR_PREAUTH_EXPIRED (Pre-authentication has expired)",
    91: "KDC_ERR_MORE_PREAUTH_DATA_REQUIRED (Additional pre-auth required)",
    92: (
        "KDC_ERR_PREAUTH_BAD_AUTHENTICATION_SET "
        "(KDC cannot accommodate requested pre-authentication element)"),
    93: "KDC_ERR_UNKNOWN_CRITICAL_FAST_OPTIONS (Unknown critical option)"
}


def parse_krb_error(data: bytes):
    if data[0] != 0x7e:
        return

    code = -1
    pos = 2 if data[1] < 128 else 2 + (data[1] & 0x7f)
    pos += 2 if data[pos+1] < 128 else 2 + (data[pos+1] & 0x7f)

    while pos < len(data):
        tag = data[pos]
        length = data[pos+1]
        if tag == 0xa6:
            code = data[pos + 4]
            if code == KDC_ERR_PREAUTH_REQUIRED:
                return
            break
        pos += 2 + length

    raise KerberosErr(
        KRB_ERRORS.get(code, f"Unknown Kerberos Error: {code}")
    )



import struct
from typing import Union


def asn1_tag(tag_num: int, content: bytes) -> bytes:
    """ASN.1 DER tagger."""
    length = len(content)
    if length < 128:
        len_octet = struct.pack('B', length)
    else:
        len_bytes = length.to_bytes((length.bit_length() + 7) // 8, 'big')
        len_octet = struct.pack('B', 0x80 + len(len_bytes)) + len_bytes
    return struct.pack('B', 0xa0 | tag_num) + len_octet + content


def asn1_len(n_or_b: Union[bytes, int]) -> bytes:
    n = n_or_b if isinstance(n_or_b, int) else len(n_or_b)
    if n <= 127:
        return struct.pack('B', n)
    l_bytes = []
    while n > 0:
        l_bytes.insert(0, n & 0xFF)
        n >>= 8
    return struct.pack('B', 0x80 | len(l_bytes)) + bytes(l_bytes)


def asn1_seq(content: bytes) -> bytes:
    return b'\x30' + asn1_len(content) + content


def asn1_gs(val: bytes) -> bytes:
    return b'\x1b' + asn1_len(val) + val


def asn1_int(val: int) -> bytes:
    num_bytes = (val.bit_length() + 7) // 8 or 1
    int_bytes = val.to_bytes(num_bytes, 'big')

    if int_bytes[0] & 0x80:
        int_bytes = b'\x00' + int_bytes

    return b'\x02' + asn1_len(int_bytes) + int_bytes


def asn1_gt(val: bytes) -> bytes:
    return b'\x18' + asn1_len(val) + val


def asn1_ostr(val: bytes) -> bytes:
    return b'\x04' + asn1_len(val) + val


def get_asn1_len(data: bytes, pos: int) -> tuple[int, int]:
    """Read ASN.1 lengte from data."""
    b = data[pos]
    if b < 128:
        return b, 1
    num_bytes = b & 0x7f
    return int.from_bytes(data[pos+1: pos+1+num_bytes], 'big'), num_bytes + 1



"""OBJREF_CUSTUM

See: 2.2.18.6, OBJREF_CUSTOM
"""
import struct
from ..dcom_const import CLSID_ActivationPropertiesIn
from ..dcom_const import IID_IActivationPropertiesIn
from ..dcom_const import FLAGS_OBJREF_CUSTOM
from ..uuid import uuid_part, bin_to_str, CLSID_SZ
from .objref import ObjRef


class ObjRefCustom(ObjRef):

    CUSTOM_FMT = '<LL'
    CUSTOM_FMT_SZ = struct.calcsize(CUSTOM_FMT)

    @classmethod
    def init(cls):
        objref = cls()

        objref.signature = 0x574F454D  # <L (1464812877)
        objref.flags = FLAGS_OBJREF_CUSTOM  # <L
        objref.iid = uuid_part(IID_IActivationPropertiesIn)  # 16b
        objref.clsid = CLSID_ActivationPropertiesIn  # 16b
        objref.cb_extension = 0  # <L
        objref.object_reference_size = None  # <L  (376)
        objref.object_data = None

        return objref

    @classmethod
    def from_data(cls, data: bytes, offset: int, size: int) -> 'ObjRefCustom':
        end = offset+size
        self = cls()

        self.read_objref(data, offset)
        offset += cls.OBJREF_SZ

        self.clsid = bin_to_str(data, offset=offset)
        offset += CLSID_SZ  # clsid size

        (
            self.cb_extension,
            self.object_reference_size
        ) = struct.unpack_from(cls.CUSTOM_FMT, data, offset=offset)
        offset += cls.CUSTOM_FMT_SZ

        self.object_data = data[offset:end]
        return self

    def set_object(self, blob: bytes):
        self.object_data = blob
        self.object_reference_size = len(blob) + 8

    def get_data(self) -> bytes:
        return super().get_data() + struct.pack(
            self.CUSTOM_FMT,
            self.cb_extension,
            self.object_reference_size
        ) + self.object_data



import struct


class DualStringArray:

    __slots__ = (
        'w_num_entries',
        'w_security_offset',
        'string_binding',
        'sec_binding',
    )

    DUALSTRING_FMT = '<HH'

    DUALSTRING_FMT_SZ = struct.calcsize(DUALSTRING_FMT)

    @classmethod
    def from_data(cls, data: bytes, offset: int) -> 'DualStringArray':
        self = cls()

        # w_num_entries: number of unsigned shorts from the first
        # entry in the StringBinding array to the end of the buffer.
        # w_security_offset: number of unsigned shorts from the first entry in
        # the StringBinding array to the first entry in the SecBinding array
        (
            cls.w_num_entries,
            cls.w_security_offset,
        ) = struct.unpack_from(cls.DUALSTRING_FMT, data, offset=offset)
        offset += cls.DUALSTRING_FMT_SZ

        end = offset + cls.w_num_entries * 2
        end_string_binding = offset + cls.w_security_offset * 2

        cls.string_binding = data[offset:end_string_binding]
        cls.sec_binding = data[end_string_binding:end]

        return self, end

    def get_data(self) -> bytes:
        return struct.pack(
            self.DUALSTRING_FMT,
            self.w_num_entries,
        ) + self.string_binding + self.sec_binding



import struct
from .common import NdrCommon
from ..dcom_const import CLSID_ActivationContextInfo


class ActivationContextInfoData(NdrCommon):

    CLSID = CLSID_ActivationContextInfo

    BUFFER32_FMT = '<llLLLL'

    @classmethod
    def get_data(cls) -> bytes:

        client_ok = 0
        reserved1 = 0
        dw_reserved1 = 0
        dw_reserved2 = 0

        # both contexts are in the format:
        #   ul_cnt_data  <L
        #   ab_data  (bytes)
        # but since we set them to 0, we just pack 0
        p_ifd_client_ctx = 0
        p_ifd_prototype_ctx = 0

        buffer = struct.pack(
            cls.BUFFER32_FMT,
            client_ok,          # <l
            reserved1,          # <l
            dw_reserved1,       # <L
            dw_reserved2,       # <L
            p_ifd_client_ctx,       # <L
            p_ifd_prototype_ctx,    # <L
        )

        private, padding = cls.get_private(buffer)

        return cls.get_common() + private + buffer + padding



"""

\x01\x10\x08\x00\xcc\xcc\xcc\xcc
\x10\x00\x00\x00\xcc\xcc\xcc\xcc


\x00\x00\x00\x00
\x00\x00\x00\x00
\x00\x00\x00\x00
\x00\x00\x00\x00'

"""
import struct
from .common import NdrCommon
from ..dcom_const import CLSID_ServerLocationInfo


class LocationInfoData(NdrCommon):

    CLSID = CLSID_ServerLocationInfo

    BUFFER_FMT = "<LLLL"

    @classmethod
    def get_data(cls) -> bytes:

        machine_name = 0
        process_id = 0
        apartment_id = 0
        context_id = 0

        buffer = struct.pack(
            cls.BUFFER_FMT,
            machine_name,       # <l
            process_id,         # <l
            apartment_id,       # <L
            context_id,         # <L
        )

        private, padding = cls.get_private(buffer)

        return cls.get_common() + private + buffer + padding



from __future__ import annotations
from typing import TYPE_CHECKING
import struct
from .common import RpcCommon
from ..tools import pad4
from .result import RpcResult
from .auth_verifier_co import RpcAuthVerifierCo
from .baseresp import RpcBaseResp

if TYPE_CHECKING:
    from ..dcom import Dcom

# /* Alter Context Response fields */
# u_int16 max_xmit_frag;    /* 16:02 max transmit frag size */
# u_int16 max_recv_frag;    /* 18:02 max receive frag size */
# u_int32 assoc_group_id;   /* 20:04 returned assoc_group_id */
# /* NOTE: NO secondary address field in AlterCtxR! */
# u_int8 pad1[4];           /* restore 4-octet alignment */
# p_result_list_t p_result_list;


class RpcAlterCtxR(RpcBaseResp):

    __slots__ = (
        'max_xmit_frag',
        'max_recv_frag',
        'assoc_group_id',
        'auth',
        'rpc_common')

    ALTER_CTX_R_FMT = '<HHL'
    ALTER_CTX_R_FMT_SIZE = struct.calcsize(ALTER_CTX_R_FMT)

    def __init__(self, dcom: Dcom, rpc_common: RpcCommon, data: bytes):
        offset = RpcCommon.COMMON_SIZE
        (
            self.max_xmit_frag,
            self.max_recv_frag,
            self.assoc_group_id
        ) = struct.unpack_from(self.ALTER_CTX_R_FMT, data, offset)

        offset += self.ALTER_CTX_R_FMT_SIZE
        offset += pad4(offset)

        n_results, reserved, reserved2 = struct.unpack_from(
            RpcResult.RESULT_LIST_FMT,
            data,
            offset
        )
        offset += RpcResult.RESULT_LIST_FMT_SIZE

        for _ in range(n_results):
            offset += 24

        offset += pad4(offset)

        if rpc_common.auth_length:
            # negTokenResp
            self.auth = RpcAuthVerifierCo(
                data,
                rpc_common.auth_length,
                offset=offset)
        else:
            self.auth = None

        self.rpc_common = rpc_common
        dcom.set_max_xmit_frag(self.max_xmit_frag)



import struct
from ..tools import pad
from .orpcthat import ORPCTHAT
from .objref_standard import ObjRefStandard
from .interface import NdrInterface


class NTLMLoginResponse(NdrInterface):

    FMT1_32 = '<LLL'
    FMT1_64 = '<QLL'

    FMT1_32_SZ = struct.calcsize(FMT1_32)

    def __init__(self, data: bytes):
        self.orpcthat, offset = ORPCTHAT.from_data(data, offset=0)

        # activation_blobs

        (
            self.referent_id,
            _,
            size
        ) = struct.unpack_from(self.FMT1_32, data, offset)
        offset += self.FMT1_32_SZ

        self.objref = ObjRefStandard.from_data(data, offset, size)
        offset += size + pad(size)

        self.error_code, = struct.unpack_from('<L', data, offset)
        assert self.error_code == 0, f'error code: {self.error_code}'

    def get_ipid(self):
        return self.objref.ipid



import struct
from typing import Tuple, Optional
from .encoded_string import EncodedString
from .class_part import ClassPart
from .heap import Heap
from .qualifier_set import QualifierSet


class ObjectBlock:

    FMT = '<L'
    FMT_SZ = struct.calcsize(FMT)

    INSTANCE = '<LBL'
    INSTANCE_SZ = struct.calcsize(INSTANCE)

    @classmethod
    def from_data(
            cls,
            data: bytes,
            offset: int,
            class_part: Optional[ClassPart] = None
            ) -> Tuple['ObjectBlock', int]:
        self = cls()

        # Start Object Block
        self.flags = data[offset]
        offset += 1

        if self.flags & 0x4:
            # WMIO - 2.2.6 - 0x4 If this flag is set,
            # the object has a Decoration block.
            self.dec_server_name, offset = \
                EncodedString.from_data(data, offset)

            self.dec_namespace_name, offset = \
                EncodedString.from_data(data, offset)

        if self.flags & 0x1:
            # The object is a CIM class.
            pass
            assert 0

        else:
            if self.flags & 0x10:
                # If this flag is set, the object is a prototype of the result
                # object for the query,
                # as specified in [MS-WMI] (section 2.2.4.1).
                # This flag MUST be used only in combination with the 0x1 flag.
                # This flag MUST NOT be used when returning IWbemClassObject,
                # which is not represented as a Prototype Result Object
                pass

            # Instance Type
            offset = self._instance_type(data, offset, class_part)

        return self, offset

    def _instance_type(
            self,
            data: bytes,
            offset: int,
            class_part: Optional[ClassPart] = None):
        if class_part:
            self.class_part = class_part
        else:
            self.class_part, offset = ClassPart.from_data(data, offset)

        # Instance
        (
            encoding_length,
            instance_flags,
            instance_class_name_n,
        ) = struct.unpack_from(self.INSTANCE, data, offset=offset)
        offset += self.INSTANCE_SZ

        # NdValueTable
        start = offset
        offset += self.class_part.nd_value_table_length
        self.nd_value_table = data[start: offset]

        # InstanceQualifierSet
        self.instance_qualifier_set, offset = \
            QualifierSet.from_data(data, offset)

        # InstPropQualSet
        flags = data[offset]
        offset += 1
        if flags & 0x2:
            assert 0, 'not supported yet'
        else:
            pass

        # InstanceHeap
        self.instance_heap, offset = Heap.from_data(data, offset)

        self.instance_qualifier_set.load(self.instance_heap)

        return offset



import struct
import asyncio
from ..logger import logger


async def send_kerberos_packet(packet_bytes: bytes,
                               kdc_host: str, kdc_port: int = 88):
    logger.debug(f"Connecting to {kdc_host}:{kdc_port}...")
    reader, writer = await asyncio.open_connection(kdc_host, kdc_port)

    try:
        tcp_packet = struct.pack('>I', len(packet_bytes)) + packet_bytes
        # print(tcp_packet)
        # logger.debug(f"Sending {len(tcp_packet)} bytes...")
        writer.write(tcp_packet)
        await writer.drain()

        header = await reader.readexactly(4)
        resp_len = struct.unpack('>I', header)[0]
        logger.debug(f"KDC anwer of {resp_len} bytes.")

        # reading next packet...
        response_data = await reader.readexactly(resp_len)
        return response_data

    except Exception as e:
        logger.error(f"Error writing to kdc: {e}")
        return None
    finally:
        writer.close()
        await writer.wait_closed()



import hashlib
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from .as_req import build_as_req, build_full_as_req
from .kdc import send_kerberos_packet
from .asn1 import get_asn1_len
from .tools import parse_krb_error


def extract_kerberos_salt(error_data: bytes) -> str:
    if error_data[0] != 0x7e:
        raise ValueError("Invalid KRB-ERROR packet")

    # find start of container
    e_data_idx = error_data.find(b'\xac')
    if e_data_idx == -1:
        raise ValueError("e-data (Tag 12) niet gevonden")

    pos = e_data_idx
    _, len_size = get_asn1_len(error_data, pos + 1)
    pos += 1 + len_size

    salt_tag_idx = error_data.find(b'\x1b\x1c', pos)
    if salt_tag_idx == -1:
        raise ValueError("Salt (GeneralString) not found")

    salt_len, salt_len_size = get_asn1_len(error_data, salt_tag_idx + 1)
    salt_start = salt_tag_idx + 1 + salt_len_size

    salt_bytes = error_data[salt_start: salt_start + salt_len]
    return salt_bytes.decode('utf-8')


def nfold_for_kerberos():
    return b'\x6b\x65\x72\x62\x65\x72\x6f\x73\x7b\x9b\x5b\x2b\x93\x13\x2b\x93'


def aes_string_to_key(password, salt):
    tkey = hashlib.pbkdf2_hmac('sha1',
                               password.encode(),
                               salt.encode(),
                               4096,
                               32)
    constant = nfold_for_kerberos()

    cipher = Cipher(algorithms.AES(tkey), modes.CBC(b'\x00' * 16))

    encryptor1 = cipher.encryptor()
    part1 = encryptor1.update(constant) + encryptor1.finalize()

    encryptor2 = cipher.encryptor()
    part2 = encryptor2.update(part1) + encryptor2.finalize()

    return part1 + part2


async def get_tgt(username: str, password: str, domain: str,
                  kdc_host: str, kdc_port: int = 88) -> tuple[bytes, bytes]:
    as_req = build_as_req(username, domain)
    resp = await send_kerberos_packet(as_req, kdc_host, kdc_port)
    parse_krb_error(resp)
    salt = extract_kerberos_salt(resp)
    base_key = aes_string_to_key(password, salt)
    full_as_req = build_full_as_req(username, domain, base_key)
    as_res_bytes = await send_kerberos_packet(full_as_req, kdc_host, kdc_port)
    parse_krb_error(as_res_bytes)
    return as_res_bytes, base_key



"""ORPCTHAT

See MS-DCOM, 2.2.13.4
"""
import struct
from typing import Tuple


class ORPCTHAT:

    FMT32 = '<LL'
    FMT64 = '<LQ'

    FMT32_SZ = struct.calcsize(FMT32)
    FMT64_SZ = struct.calcsize(FMT64)

    EXT = '<LL'

    EXT_SZ = struct.calcsize(EXT)

    EXT32 = '<LL'
    EXT64 = '<LL'

    EXT32_SZ = struct.calcsize(EXT32)
    EXT64_SZ = struct.calcsize(EXT64)

    @classmethod
    def from_data(cls, data: bytes, offset: int = 0) -> Tuple['ORPCTHAT', int]:
        orpcthat = cls()

        orpcthat.flags, extensions = struct.unpack_from(
            cls.FMT32,
            data,
            offset=offset)
        offset += cls.FMT32_SZ

        if extensions:
            assert 0, 'extensions are not implemented correctly (yet)'
            (
                size,
                reserved,
            ) = struct.unpack_from(cls.EXT, data, offset=offset)
            offset += cls.EXT_SZ
            for _ in range(size):
                referent_id, = struct.unpack_from('<L', data, offset=offset)
                offset += 4  # referent_id
                if referent_id:
                    offset += 16  # GUID
                    sz, = struct.unpack_from('<L', data, offset=offset)
                    offset += 4  # What to do with sz?

        return orpcthat, offset



"""OBJREF_CUSTUM

See: 2.2.18.6, OBJREF_CUSTOM
"""
import struct
from typing import Union, TYPE_CHECKING
from ..uuid import bin_to_str, CLSID_SZ
from .const import FLAGS_OBJREF_CUSTOM
from .const import FLAGS_OBJREF_EXTENDED
from .const import FLAGS_OBJREF_STANDARD

if TYPE_CHECKING:
    from .objref_standard import ObjRefStandard
    from .objref_custom import ObjRefCustom


class ObjRef:

    OBJREF_FMT = '<LL'
    OBJREF_FMT_SZ = struct.calcsize(OBJREF_FMT)

    OBJREF_SZ = OBJREF_FMT_SZ + CLSID_SZ

    def read_objref(self, data: bytes, offset: int):
        (
            self.signature,
            self.flags,
        ) = struct.unpack_from(self.OBJREF_FMT, data, offset=offset)
        offset += self.OBJREF_FMT_SZ

        self.iid = bin_to_str(data, offset=offset)
        offset += CLSID_SZ  # iid size

    def get_data(self) -> bytes:
        return struct.pack(
            self.OBJREF_FMT,
            self.signature,
            self.flags,
        ) + self.iid + self.clsid

    @classmethod
    def from_data(cls, data: bytes, offset: int, size: int) \
            -> Union['ObjRefStandard', 'ObjRefCustom']:
        (
            signature,
            flags,
        ) = struct.unpack_from(cls.OBJREF_FMT, data, offset=offset)

        if flags & FLAGS_OBJREF_STANDARD:
            from .objref_standard import ObjRefStandard
            return ObjRefStandard.from_data(data, offset, size)

        if flags & FLAGS_OBJREF_CUSTOM:
            from .objref_custom import ObjRefCustom
            return ObjRefCustom.from_data(data, offset, size)

        if flags & FLAGS_OBJREF_EXTENDED:
            raise NotImplementedError('OBJREF_EXTENDED not implemented yet')
            from .objref_extended import ObjRefExtended
            return ObjRefExtended.from_data(data, offset, size)

        assert 0, f'unsupported ObjRef (flags: {flags})'



import random
from .asn1 import (
    asn1_len, asn1_tag, asn1_seq, asn1_gs, asn1_ostr, asn1_int, asn1_gt
)
from .kdc import send_kerberos_packet
from .tools import decrypt_kerberos_aes_cts, encrypt_kerberos_aes_cts
from .tools import read_session_key, parse_krb_error
from .asn1 import get_asn1_len
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from datetime import datetime, timezone
from pyasn1.codec.der import decoder, encoder
from pyasn1.type import univ, tag, namedtype, char


def aes_cts_encrypt(key, plaintext):
    iv = b'\x00' * 16
    n = len(plaintext)

    if n % 16 == 0:
        cipher = Cipher(algorithms.AES(key), modes.CBC(iv))
        encryptor = cipher.encryptor()
        return encryptor.update(plaintext) + encryptor.finalize()

    pad_len = 16 - (n % 16)
    padded_plain = plaintext + b'\x00' * pad_len

    cipher = Cipher(algorithms.AES(key), modes.CBC(iv))
    encryptor = cipher.encryptor()
    full_cipher = encryptor.update(padded_plain) + encryptor.finalize()

    last_block_start = (len(full_cipher) // 16 - 1) * 16
    prev_block_start = last_block_start - 16

    c_prev = full_cipher[prev_block_start:last_block_start]
    c_last = full_cipher[last_block_start:]

    return full_cipher[:prev_block_start] + c_last + c_prev[:n % 16]


class AS_REP(univ.Sequence):
    tagSet = univ.Sequence.tagSet.tagExplicitly(
        tag.Tag(tag.tagClassApplication, tag.tagFormatConstructed, 11)
    )

    componentType = namedtype.NamedTypes(
        namedtype.NamedType('pvno', univ.Integer().subtype(
            explicitTag=tag.Tag(tag.tagClassContext,
                                tag.tagFormatConstructed, 0))),
        namedtype.NamedType('msg-type', univ.Integer().subtype(
            explicitTag=tag.Tag(tag.tagClassContext,
                                tag.tagFormatConstructed, 1))),
        namedtype.OptionalNamedType(
            'padata',
            univ.SequenceOf(univ.Sequence()).subtype(
                explicitTag=tag.Tag(tag.tagClassContext,
                                    tag.tagFormatConstructed, 2))),
        namedtype.NamedType('crealm', char.GeneralString().subtype(
            explicitTag=tag.Tag(tag.tagClassContext,
                                tag.tagFormatConstructed, 3))),
        namedtype.NamedType('cname', univ.Any().subtype(
            explicitTag=tag.Tag(tag.tagClassContext,
                                tag.tagFormatConstructed, 4))),
        namedtype.NamedType('ticket', univ.Any().subtype(
            explicitTag=tag.Tag(tag.tagClassContext,
                                tag.tagFormatConstructed, 5))),
        namedtype.NamedType('enc-part', univ.Any().subtype(
            explicitTag=tag.Tag(tag.tagClassContext,
                                tag.tagFormatConstructed, 6)))
    )


class EncryptedData(univ.Sequence):
    componentType = namedtype.NamedTypes(
        namedtype.NamedType('etype', univ.Integer().subtype(
            explicitTag=tag.Tag(tag.tagClassContext,
                                tag.tagFormatConstructed, 0))),
        namedtype.OptionalNamedType('kvno', univ.Integer().subtype(
            explicitTag=tag.Tag(tag.tagClassContext,
                                tag.tagFormatConstructed, 1))),
        namedtype.NamedType('cipher', univ.OctetString().subtype(
            explicitTag=tag.Tag(tag.tagClassContext,
                                tag.tagFormatConstructed, 2)))
    )


def get_session_key(as_rep_bytes: bytes, base_key: bytes) -> bytes:
    as_rep_obj, _ = decoder.decode(as_rep_bytes, asn1Spec=AS_REP())

    enc_part_any = as_rep_obj['enc-part']
    enc_data_obj, _ = decoder.decode(
        bytes(enc_part_any),
        asn1Spec=EncryptedData()
    )

    cipher_blob = bytes(enc_data_obj['cipher'])

    decrypted = decrypt_kerberos_aes_cts(base_key, 3, cipher_blob)
    session_key = read_session_key(decrypted)

    return session_key


def build_tgs_req(username: str,
                  domain: str,
                  session_key: bytes,
                  ticket_bytes: bytes,
                  target_service: tuple[str, str]):
    now = datetime.now(timezone.utc)
    ts_str = now.strftime("%Y%m%d%H%M%SZ").encode()

    cname_content = (
        asn1_tag(0, asn1_int(1)) +
        asn1_tag(1, asn1_seq(asn1_gs(username.encode())))
    )

    domcaps = domain.upper()
    auth_body = (
        asn1_tag(0, asn1_int(5)) +                       # pvno
        asn1_tag(1, asn1_gs(domcaps.encode())) +         # crealm
        asn1_tag(2, asn1_seq(cname_content)) +           # cname
        asn1_tag(4, asn1_int(now.microsecond)) +         # cusec
        asn1_tag(5, asn1_gt(ts_str))                     # ctime
    )

    # Application Tag 2 (0x62) -> Authenticator
    auth_plain = b'\x62' + asn1_len(asn1_seq(auth_body)) + asn1_seq(auth_body)

    # Encrypt Authenticator (7 -> TGS-REQ) ---
    final_cipher = encrypt_kerberos_aes_cts(session_key, 7, auth_plain)
    auth_enc_seq = asn1_seq(
        asn1_tag(0, asn1_int(18)) +                      # etype AES256
        asn1_tag(2, asn1_ostr(final_cipher))               # cipher
    )

    # AP-REQ (-> PA-DATA)
    ap_req_body = (
        asn1_tag(0, asn1_int(5)) +                       # pvno
        asn1_tag(1, asn1_int(14)) +                      # msg-type (14=AP-REQ)
        asn1_tag(2, b'\x03\x05\x00\x00\x00\x00\x00') +   # ap-options
        asn1_tag(3, ticket_bytes) +                      # ticket
        asn1_tag(4, auth_enc_seq)                        # authenticator
    )
    ap_req_seq = asn1_seq(ap_req_body)
    encoded_ap_req = b'\x6e' + asn1_len(ap_req_seq) + ap_req_seq

    padata_item = asn1_seq(
        asn1_tag(1, asn1_int(1)) +                       # PA-TGS-REQ
        asn1_tag(2, asn1_ostr(encoded_ap_req))           # value
    )
    padata_field = asn1_tag(3, asn1_seq(padata_item))    # [3] padata

    service, host_fqdn = target_service
    sname_inner = (
        asn1_tag(0, asn1_int(2)) +                       # NT-SRV-INST
        asn1_tag(1, asn1_seq(
            asn1_gs(service.encode()) +
            asn1_gs(host_fqdn.encode())
        ))
    )

    # Etype list: RC4, AES128, DES, AES256
    etype_list = asn1_int(23) + asn1_int(16) + asn1_int(3) + asn1_int(18)
    nonce = random.getrandbits(31)

    req_body_content = (
        asn1_tag(0, b'\x03\x05\x00\x40\x81\x00\x10') +   # KDC Options
        asn1_tag(2, asn1_gs(domcaps.encode())) +         # Realm
        asn1_tag(3, asn1_seq(sname_inner)) +             # sname
        asn1_tag(5, asn1_gt(ts_str)) +                   # Till
        asn1_tag(7, asn1_int(nonce)) +                   # Nonce
        asn1_tag(8, asn1_seq(etype_list))                # Etypes
    )
    req_body_field = asn1_tag(4, asn1_seq(req_body_content))

    # TGS-REQ
    final_content = (
        asn1_tag(1, asn1_int(5)) +                       # pvno
        asn1_tag(2, asn1_int(12)) +                      # msg-type 12=TGS-REQ
        padata_field +
        req_body_field
    )

    # Application Tag 12 (0x6c) -> TGS-REQ
    inner_seq = asn1_seq(final_content)
    return b'\x6c' + asn1_len(inner_seq) + inner_seq


def extract_ticket(data):
    pos = 0
    if data[pos] not in [0x6b, 0x6d]:
        raise ValueError("Invalid Kerberos answer")

    _, len_size = get_asn1_len(data, pos + 1)
    pos += 1 + len_size  # AS-REP

    # Skip SEQUENCE tag (0x30)
    if data[pos] != 0x30:
        raise ValueError("SEQUENCE expected")
    _, len_size = get_asn1_len(data, pos + 1)
    pos += 1 + len_size

    while pos < len(data):
        tag = data[pos]
        content_len, len_size = get_asn1_len(data, pos + 1)

        if tag == 0xa5:
            ticket_start = pos + 1 + len_size
            tkt_len, tkt_len_size = get_asn1_len(data, ticket_start + 1)
            total_tkt_size = 1 + tkt_len_size + tkt_len
            return data[ticket_start: ticket_start + total_tkt_size]

        pos += 1 + len_size + content_len

    raise ValueError("Ticket tag [5] not found in sequence")


def get_service_key(resp_bytes: bytes,
                    session_key: bytes) -> tuple[bytes, bytes]:
    raw_obj, _ = decoder.decode(resp_bytes)
    enc_part = raw_obj.getComponentByPosition(5)

    cipher_blob = None
    for i in range(len(enc_part)):
        comp = enc_part.getComponentByPosition(i)
        if isinstance(comp, univ.OctetString):
            cipher_blob = bytes(comp)
            break

    if not cipher_blob:
        cipher_blob = bytes(enc_part[2])

    decrypted_raw = decrypt_kerberos_aes_cts(session_key, 8, cipher_blob)
    payload = decrypted_raw[16:]
    inner_obj, _ = decoder.decode(payload)

    key_sequence = inner_obj.getComponentByPosition(0)
    service_session_key = bytes(key_sequence.getComponentByPosition(1))

    ticket = raw_obj.getComponentByPosition(4)
    ticket_bytes = encoder.encode(ticket)

    WRAPPERS = [0xa3, 0xa4, 0xa5]

    if ticket_bytes[0] in WRAPPERS:
        try:
            ticket_start = ticket_bytes.index(b'\x61', 0, 10)
            ticket_bytes = ticket_bytes[ticket_start:]
        except ValueError:
            pass

    return ticket_bytes, service_session_key


async def get_tgs(username: str, domain: str, host: str,
                  as_rep_bytes: bytes, base_key: bytes,
                  kdc_host: str, kdc_port: int = 88) -> bytes:
    tgs_session_key = get_session_key(as_rep_bytes, base_key)
    tgs_ticket = extract_ticket(as_rep_bytes)
    tgs_req = build_tgs_req(username,
                            domain,
                            tgs_session_key,
                            tgs_ticket,
                            ("host", host))
    as_res_bytes = await send_kerberos_packet(tgs_req, kdc_host, kdc_port)
    parse_krb_error(as_res_bytes)
    ticket, service_key = get_service_key(as_res_bytes, tgs_session_key)
    return ticket, service_key



NTLM_AUTH_NONE = 1
NTLM_AUTH_CONNECT = 2
NTLM_AUTH_CALL = 3
NTLM_AUTH_PKT = 4
NTLM_AUTH_PKT_INTEGRITY = 5
NTLM_AUTH_PKT_PRIVACY = 6

# If set, requests 56-bit encryption. If the client sends NTLMSSP_
# NEGOTIATE_SEAL or NTLMSSP_NEGOTIATE_SIGN
# with NTLMSSP_NEGOTIATE_56 to the server in the NEGOTIATE_MESSAGE, the server
# MUST return NTLMSSP_NEGOTIATE_56 to
# the client in the CHALLENGE_MESSAGE. Otherwise it is ignored. If both
# NTLMSSP_NEGOTIATE_56 and NTLMSSP_NEGOTIATE_128
# are requested and supported by the client and server, NTLMSSP_NEGOTIATE_56
# and NTLMSSP_NEGOTIATE_128 will both be
# returned to the client. Clients and servers that set NTLMSSP_NEGOTIATE_SEAL
# SHOULD set NTLMSSP_NEGOTIATE_56 if it is
# supported. An alternate name for this field is NTLMSSP_NEGOTIATE_56.
NTLMSSP_NEGOTIATE_56 = 0x80000000

# If set, requests an explicit key exchange. This capability SHOULD be used
# because it improves security for message
# integrity or confidentiality. See sections 3.2.5.1.2, 3.2.5.2.1,
# and 3.2.5.2.2 for details. An alternate name for
# this field is NTLMSSP_NEGOTIATE_KEY_EXCH.
NTLMSSP_NEGOTIATE_KEY_EXCH = 0x40000000

# If set, requests 128-bit session key negotiation. An alternate name for this
# field is NTLMSSP_NEGOTIATE_128.
# If the client sends NTLMSSP_NEGOTIATE_128 to the server in
# the NEGOTIATE_MESSAGE, the server MUST return
# NTLMSSP_NEGOTIATE_128 to the client in the CHALLENGE_MESSAGE only if the
# client sets NTLMSSP_NEGOTIATE_SEAL or
# NTLMSSP_NEGOTIATE_SIGN. Otherwise it is ignored.
# If both NTLMSSP_NEGOTIATE_56 and NTLMSSP_NEGOTIATE_128 are
# requested and supported by the client and server, NTLMSSP_NEGOTIATE_56 and
# NTLMSSP_NEGOTIATE_128 will both be
# returned to the client. Clients and servers that set NTLMSSP_NEGOTIATE_SEAL
# SHOULD set NTLMSSP_NEGOTIATE_128 if it
# is supported. An alternate name for this field is NTLMSSP_NEGOTIATE_128
NTLMSSP_NEGOTIATE_128 = 0x20000000

NTLMSSP_RESERVED_1 = 0x10000000
NTLMSSP_RESERVED_2 = 0x08000000
NTLMSSP_RESERVED_3 = 0x04000000

# If set, requests the protocol version number. The data corresponding to this
# flag is provided in the Version field
# of the NEGOTIATE_MESSAGE, the CHALLENGE_MESSAGE, and
# the AUTHENTICATE_MESSAGE. <22> An alternate name for this field
# is NTLMSSP_NEGOTIATE_VERSION
NTLMSSP_NEGOTIATE_VERSION = 0x02000000
NTLMSSP_RESERVED_4 = 0x01000000

# If set, indicates that the TargetInfo fields in the
# CHALLENGE_MESSAGE (section 2.2.1.2) are populated.
# An alternate name for this field is NTLMSSP_NEGOTIATE_TARGET_INFO.
NTLMSSP_NEGOTIATE_TARGET_INFO = 0x00800000

# If set, requests the usage of the LMOWF (section 3.3). An alternate name
# for this field is
# NTLMSSP_REQUEST_NON_NT_SESSION_KEY.
NTLMSSP_REQUEST_NON_NT_SESSION_KEY = 0x00400000
NTLMSSP_RESERVED_5 = 0x00200000

# If set, requests an identify level token. An alternate name for this field
# is NTLMSSP_NEGOTIATE_IDENTIFY
NTLMSSP_NEGOTIATE_IDENTIFY = 0x00100000

# If set, requests usage of the NTLM v2 session security. NTLM v2 session
# security is a misnomer because it is not
# NTLM v2. It is NTLM v1 using the extended session security that is also
# in NTLM v2. NTLMSSP_NEGOTIATE_LM_KEY and
# NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY are mutually exclusive. If
# both NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY
# and NTLMSSP_NEGOTIATE_LM_KEY are requested,
#  NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY alone MUST be returned to the
# client. NTLM v2 authentication session key generation MUST be supported by
# both the client and the DC in order to be
# used, and extended session security signing and sealing requires support
# from the client and the server in order to
# be used.<23> An alternate name for this field is
# NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY
NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY = 0x00080000
NTLMSSP_NEGOTIATE_NTLM2 = 0x00080000
NTLMSSP_TARGET_TYPE_SHARE = 0x00040000

# If set, TargetName MUST be a server name. The data corresponding to this
# flag is provided by the server in the
# TargetName field of the CHALLENGE_MESSAGE. If this bit is set, then
# NTLMSSP_TARGET_TYPE_DOMAIN MUST NOT be set.
# This flag MUST be ignored in the NEGOTIATE_MESSAGE and the
# AUTHENTICATE_MESSAGE. An alternate name for this field
# is NTLMSSP_TARGET_TYPE_SERVER
NTLMSSP_TARGET_TYPE_SERVER = 0x00020000

# If set, TargetName MUST be a domain name. The data corresponding to this
# flag is provided by the server in the
# TargetName field of the CHALLENGE_MESSAGE. If set, then
# NTLMSSP_TARGET_TYPE_SERVER MUST NOT be set. This flag MUST
# be ignored in the NEGOTIATE_MESSAGE and the AUTHENTICATE_MESSAGE.
# An alternate name for this field is
# NTLMSSP_TARGET_TYPE_DOMAIN.
NTLMSSP_TARGET_TYPE_DOMAIN = 0x00010000

# If set, requests the presence of a signature block on all messages.
# NTLMSSP_NEGOTIATE_ALWAYS_SIGN MUST be set in the
# NEGOTIATE_MESSAGE to the server and the CHALLENGE_MESSAGE to the client.
# NTLMSSP_NEGOTIATE_ALWAYS_SIGN is overridden
# by NTLMSSP_NEGOTIATE_SIGN and NTLMSSP_NEGOTIATE_SEAL, if they are supported.
# An alternate name for this field is
# NTLMSSP_NEGOTIATE_ALWAYS_SIGN.
# forces the other end to sign packets
NTLMSSP_NEGOTIATE_ALWAYS_SIGN = 0x00008000
NTLMSSP_RESERVED_6 = 0x00004000

# This flag indicates whether the Workstation field is present. If this flag
# is not set, the Workstation field MUST be
# ignored. If this flag is set, the length field of the Workstation field
# specifies whether the workstation name is
# nonempty or not.<24> An alternate name for this field is
# NTLMSSP_NEGOTIATE_OEM_WORKSTATION_SUPPLIED.
NTLMSSP_NEGOTIATE_OEM_WORKSTATION_SUPPLIED = 0x00002000

# If set, the domain name is provided (section 2.2.1.1).<25> An alternate name
# for this field is
# NTLMSSP_NEGOTIATE_OEM_DOMAIN_SUPPLIED
NTLMSSP_NEGOTIATE_OEM_DOMAIN_SUPPLIED = 0x00001000

# If set, the connection SHOULD be anonymous
NTLMSSP_NEGOTIATE_ANONYMOUS = 0x00000800

# If set, LM authentication is not allowed and only NT authentication is used.
NTLMSSP_NEGOTIATE_NT_ONLY = 0x00000400

# If set, requests usage of the NTLM v1 session security protocol.
# NTLMSSP_NEGOTIATE_NTLM MUST be set in the
# NEGOTIATE_MESSAGE to the server and the CHALLENGE_MESSAGE to the client.
# An alternate name for this field is
# NTLMSSP_NEGOTIATE_NTLM
NTLMSSP_NEGOTIATE_NTLM = 0x00000200
NTLMSSP_RESERVED_8 = 0x00000100

# If set, requests LAN Manager (LM) session key computation.
# NTLMSSP_NEGOTIATE_LM_KEY and
# NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY are mutually exclusive. If both
# NTLMSSP_NEGOTIATE_LM_KEY and
# NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY are requested,
# NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY alone MUST be
# returned to the client. NTLM v2 authentication session key generation MUST
# be supported by both the client and the
# DC in order to be used, and extended session security signing and sealing
# requires support from the client and the
# server to be used. An alternate name for this field is
# NTLMSSP_NEGOTIATE_LM_KEY.
NTLMSSP_NEGOTIATE_LM_KEY = 0x00000080

# If set, requests connectionless authentication.
# If NTLMSSP_NEGOTIATE_DATAGRAM is set, then NTLMSSP_NEGOTIATE_KEY_EXCH
# MUST always be set in the AUTHENTICATE_MESSAGE to the server and the
# CHALLENGE_MESSAGE to the client. An alternate
# name for this field is NTLMSSP_NEGOTIATE_DATAGRAM.
NTLMSSP_NEGOTIATE_DATAGRAM = 0x00000040

# If set, requests session key negotiation for message confidentiality.
# If the client sends NTLMSSP_NEGOTIATE_SEAL to
# the server in the NEGOTIATE_MESSAGE, the server MUST return
# NTLMSSP_NEGOTIATE_SEAL to the client in the
# CHALLENGE_MESSAGE. Clients and servers that set
# NTLMSSP_NEGOTIATE_SEAL SHOULD always set NTLMSSP_NEGOTIATE_56 and
# NTLMSSP_NEGOTIATE_128, if they are supported. An alternate name for this
# field is NTLMSSP_NEGOTIATE_SEAL.
NTLMSSP_NEGOTIATE_SEAL = 0x00000020

# If set, requests session key negotiation for message signatures. If the
# client sends NTLMSSP_NEGOTIATE_SIGN to the
# server in the NEGOTIATE_MESSAGE, the server MUST return
# NTLMSSP_NEGOTIATE_SIGN to the client in the CHALLENGE_MESSAGE.
# An alternate name for this field is NTLMSSP_NEGOTIATE_SIGN.
# means packet is signed, if verifier is wrong it fails
NTLMSSP_NEGOTIATE_SIGN = 0x00000010
NTLMSSP_RESERVED_9 = 0x00000008

# If set, a TargetName field of the CHALLENGE_MESSAGE (section 2.2.1.2) MUST
# be supplied. An alternate name for this
# field is NTLMSSP_REQUEST_TARGET.
NTLMSSP_REQUEST_TARGET = 0x00000004

# If set, requests OEM character set encoding. An alternate name for this
# field is NTLM_NEGOTIATE_OEM. See bit A for
# details.
NTLM_NEGOTIATE_OEM = 0x00000002

# If set, requests Unicode character set encoding. An alternate name for
# this field is NTLMSSP_NEGOTIATE_UNICODE.
NTLMSSP_NEGOTIATE_UNICODE = 0x00000001

# AV_PAIR constants
NTLMSSP_AV_EOL = 0x00
NTLMSSP_AV_HOSTNAME = 0x01
NTLMSSP_AV_DOMAINNAME = 0x02
NTLMSSP_AV_DNS_HOSTNAME = 0x03
NTLMSSP_AV_DNS_DOMAINNAME = 0x04
NTLMSSP_AV_DNS_TREENAME = 0x05
NTLMSSP_AV_FLAGS = 0x06
NTLMSSP_AV_TIME = 0x07
NTLMSSP_AV_RESTRICTIONS = 0x08
NTLMSSP_AV_TARGET_NAME = 0x09
NTLMSSP_AV_CHANNEL_BINDINGS = 0x0a

NTLMSSP_REVISION_W2K3 = 0x0F  # Version 15 of the NTLMSSP is in use.



import struct
import binascii
from typing import Callable
from .const import (
    NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY,
    NTLMSSP_NEGOTIATE_KEY_EXCH,
)
from ..tools import hmac_md5


class NTLMMessageSignature:

    @staticmethod
    def get_data(
            flags: int,
            seq_num: int,
            message_to_sign: bytes,
            signing_key: bytes,
            handle: Callable[[bytes, bytes], bytes]):

        version = 1

        if flags & NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY:

            checksum = hmac_md5(
                signing_key,
                struct.pack('<i', seq_num) + message_to_sign)[:8]

            if flags & NTLMSSP_NEGOTIATE_KEY_EXCH:
                checksum = handle(checksum)

            checksum, = struct.unpack('<q', checksum)

            signature = struct.pack(
                '<LqI',
                version,
                checksum,
                seq_num)
            return signature

        checksum = \
            struct.pack('<I', binascii.crc32(message_to_sign) & 0xFFFFFFFF)
        checksum, = struct.unpack('<I', handle(checksum))
        seq_num = struct.unpack('<I', handle(b'\x00\x00\x00\x00'))[0] ^ seq_num
        random_pad = 0

        signature = struct.pack(
                '<LIII',
                version,
                random_pad,
                checksum,
                seq_num)
        return signature



import struct
from datetime import datetime, timedelta
from typing import Tuple, Union


class CimType:
    CIM_ARRAY_FLAG = 0x2000
    CIM_INHERITED_FLAG = 0x4000

    CIM_TYPE_SINT8 = 16
    CIM_TYPE_UINT8 = 17
    CIM_TYPE_SINT16 = 2
    CIM_TYPE_UINT16 = 18
    CIM_TYPE_SINT32 = 3
    CIM_TYPE_UINT32 = 19
    CIM_TYPE_SINT64 = 20
    CIM_TYPE_UINT64 = 21
    CIM_TYPE_REAL32 = 4
    CIM_TYPE_REAL64 = 5
    CIM_TYPE_BOOLEAN = 11
    CIM_TYPE_STRING = 8
    CIM_TYPE_DATETIME = 101
    CIM_TYPE_REFERENCE = 102
    CIM_TYPE_CHAR16 = 103
    CIM_TYPE_OBJECT = 13
    CIM_ARRAY_SINT8 = 8208
    CIM_ARRAY_UINT8 = 8209
    CIM_ARRAY_SINT16 = 8194
    CIM_ARRAY_UINT16 = 8210
    CIM_ARRAY_SINT32 = 8195
    CIM_ARRAY_UINT32 = 8201
    CIM_ARRAY_SINT64 = 8202
    CIM_ARRAY_UINT64 = 8203
    CIM_ARRAY_REAL32 = 8196
    CIM_ARRAY_REAL64 = 8197
    CIM_ARRAY_BOOLEAN = 8203
    CIM_ARRAY_STRING = 8200
    CIM_ARRAY_DATETIME = 8293
    CIM_ARRAY_REFERENCE = 8294
    CIM_ARRAY_CHAR16 = 8295
    CIM_ARRAY_OBJECT = 8205

    CIM_TYPES_REF = {
        CIM_TYPE_SINT8: ('b', struct.calcsize('b')),
        CIM_TYPE_UINT8: ('B', struct.calcsize('B')),
        CIM_TYPE_SINT16: ('<h', struct.calcsize('<h')),
        CIM_TYPE_UINT16: ('<H', struct.calcsize('<H')),
        CIM_TYPE_SINT32: ('<l', struct.calcsize('<l')),
        CIM_TYPE_UINT32: ('<L', struct.calcsize('<L')),
        CIM_TYPE_SINT64: ('<q', struct.calcsize('<q')),
        CIM_TYPE_UINT64: ('<Q', struct.calcsize('<Q')),
        CIM_TYPE_REAL32: ('<f', struct.calcsize('<f')),
        CIM_TYPE_REAL64: ('<d', struct.calcsize('<d')),
        CIM_TYPE_BOOLEAN: ('<H', struct.calcsize('<H')),
        CIM_TYPE_STRING: ('<L', struct.calcsize('<L')),  # HEAPREF
        CIM_TYPE_DATETIME: ('<L', struct.calcsize('<L')),  # HEAPREF
        CIM_TYPE_REFERENCE: ('<L', struct.calcsize('<L')),  # HEAPREF
        CIM_TYPE_CHAR16: ('<H', struct.calcsize('<H')),
        CIM_TYPE_OBJECT: ('<L', struct.calcsize('<L')),  # HEAPREF
    }

    _CIM_TYPES_NAME = {
        CIM_TYPE_SINT8: 'sint8',
        CIM_TYPE_UINT8: 'uint8',
        CIM_TYPE_SINT16: 'sint16',
        CIM_TYPE_UINT16: 'uint16',
        CIM_TYPE_SINT32: 'sint32',
        CIM_TYPE_UINT32: 'uint32',
        CIM_TYPE_SINT64: 'sint64',
        CIM_TYPE_UINT64: 'uint64',
        CIM_TYPE_REAL32: 'real32',
        CIM_TYPE_REAL64: 'real64',
        CIM_TYPE_BOOLEAN: 'bool',
        CIM_TYPE_STRING: 'string',
        CIM_TYPE_DATETIME: 'datetime',
        CIM_TYPE_REFERENCE: 'reference',
        CIM_TYPE_CHAR16: 'char16',
        CIM_TYPE_OBJECT: 'object',
    }

    _CIM_TYPES_PYTYPE = {
        CIM_TYPE_SINT8: int,
        CIM_TYPE_UINT8: int,
        CIM_TYPE_SINT16: int,
        CIM_TYPE_UINT16: int,
        CIM_TYPE_SINT32: int,
        CIM_TYPE_UINT32: int,
        CIM_TYPE_SINT64: int,
        CIM_TYPE_UINT64: int,
        CIM_TYPE_REAL32: float,
        CIM_TYPE_REAL64: float,
        CIM_TYPE_BOOLEAN: bool,
        CIM_TYPE_STRING: str,
        CIM_TYPE_DATETIME: Union[datetime, timedelta],
        CIM_TYPE_REFERENCE: str,
        CIM_TYPE_CHAR16: str,
        CIM_TYPE_OBJECT: object,
    }

    @classmethod
    def get_value(cls, key: int, data: bytes, offset: int) -> Tuple[int, int]:
        fmt, size = cls.CIM_TYPES_REF[key]
        return struct.unpack_from(fmt, data, offset=offset)[0], offset + size

    @classmethod
    def get_cim_type_ref(cls, cim_type):
        p_type = \
            cim_type & (~(CimType.CIM_ARRAY_FLAG | CimType.CIM_INHERITED_FLAG))
        return cls.CIM_TYPES_REF[p_type]

    @classmethod
    def get_cim_type_name(cls, cim_type):
        p_type = \
            cim_type & (~(CimType.CIM_ARRAY_FLAG | CimType.CIM_INHERITED_FLAG))
        return cls._CIM_TYPES_NAME[p_type]

    @classmethod
    def get_cim_type_pytype(cls, cim_type) -> type:
        p_type = \
            cim_type & (~(CimType.CIM_ARRAY_FLAG | CimType.CIM_INHERITED_FLAG))
        return cls._CIM_TYPES_PYTYPE[p_type]



import datetime
import struct
from .asn1 import asn1_len, asn1_seq, asn1_tag, asn1_int, asn1_gt, asn1_ostr
from .tools import encrypt_kerberos_rc4
from ..dcom_const import NDR_TransferSyntaxIdentifier
from ..rpc.const import RPC_C_AUTHN_GSS_NEGOTIATE


def get_neg_token(service_session_key: bytes, seq_number: int):
    now = datetime.datetime.now(datetime.timezone.utc)
    timestamp = now.strftime('%Y%m%d%H%M%SZ').encode('ascii')
    cusec = now.microsecond

    enc_ap_rep_body = (
        asn1_tag(0, asn1_gt(timestamp)) +   # [0] ctime
        asn1_tag(1, asn1_int(cusec)) +      # [1] cusec
        asn1_tag(3, asn1_int(seq_number))
    )
    inner_seq = asn1_seq(enc_ap_rep_body)
    enc_ap_rep = b'\x7b' + asn1_len(inner_seq) + inner_seq

    encrypted_data = encrypt_kerberos_rc4(service_session_key, 12, enc_ap_rep)
    enc_part_content = (
        asn1_tag(0, asn1_int(23)) +                  # etype 23 (RC4)
        asn1_tag(2, asn1_ostr(encrypted_data))       # cipher
    )

    ap_rep_body = (
        asn1_tag(0, asn1_int(5)) +                   # pvno
        asn1_tag(1, asn1_int(15)) +                  # msg-type 15 (AP-REP)
        asn1_tag(2, asn1_seq(enc_part_content))      # enc-part
    )

    ap_rep = b'\x6f' + asn1_len(asn1_seq(ap_rep_body)) + asn1_seq(ap_rep_body)
    response_token = asn1_tag(2, asn1_ostr(ap_rep))
    auth_neg_token = asn1_tag(1, asn1_seq(response_token))

    return auth_neg_token


def build_alter_context(iid: bytes,
                        call_id: int,
                        auth_level: int,
                        context_id: int,
                        neg_token: bytes):
    ndr_ts = NDR_TransferSyntaxIdentifier

    sec_trailer = struct.pack('<BBBB',
                              RPC_C_AUTHN_GSS_NEGOTIATE,
                              auth_level,
                              0x00, 0x00) + struct.pack('<I', context_id)

    ctx_body = (
        b'\x00\x00\x00\x00' +       # Padding
        b'\x01\x00\x00\x00' +       # NumCtxItems: 1
        b'\x00\x00' +               # Context ID: 0
        b'\x01\x00' +               # NumTransItems: 1
        iid +                       # IID (20 bytes incl version)
        ndr_ts +                    # Syntax (20 bytes)
        b'\x00\x00\x01\x00' +       # Impacket indicator / marker
        iid +                       # IID repeat (20 bytes)
        ndr_ts                      # Syntax repeat (20 bytes)
    )

    auth_len = len(neg_token)
    frag_len = 124 + auth_len

    header = (
        b'\x05\x00' +               # RPC Version 5.0
        b'\x0e' +                   # PDU Type: Alter Context
        b'\x03' +                   # Flags: First + Last
        b'\x10\x00\x00\x00' +       # Data Representation
        struct.pack('<H', frag_len) +
        struct.pack('<H', auth_len) +
        struct.pack('<I', call_id) +
        b'\xb8\x10\xb8\x10'         # Max Xmit/Recv (4280)
    )

    return header + ctx_body + sec_trailer + neg_token



import struct
from .heap import Heap
from .encoded_string import EncodedString
from .qualifier_set import QualifierSet
from .properties import Properties


class ClassPart:

    HEADER = '<LBLL'
    HEADER_SZ = struct.calcsize(HEADER)

    DIRIVATION_LIST = '<L'
    DIRIVATION_LIST_SZ = struct.calcsize(DIRIVATION_LIST)

    properties: Properties

    @classmethod
    def from_data(cls, data, offset):
        keep = offset

        # ClassHeader
        self = cls()
        (
            encoding_length,
            reserved,
            self.class_name_ref,
            self.nd_value_table_length,
        ) = struct.unpack_from(cls.HEADER, data, offset=offset)
        offset += cls.HEADER_SZ

        # DerivationList
        (
            enc_length,
        ) = struct.unpack_from(cls.DIRIVATION_LIST, data, offset=offset)
        end = offset + enc_length
        start = offset + cls.DIRIVATION_LIST_SZ

        self.class_name_encoding = data[start: end]
        offset += enc_length

        # QualifierSet
        self.qualifier_set, offset = QualifierSet.from_data(data, offset)

        # Properties
        self.properties, offset = Properties.from_data(data, offset)

        start = offset
        offset += self.nd_value_table_length
        self.nd_value_table = data[start: offset]

        # ClassHeap
        self.class_heap, offset = Heap.from_data(data, offset)

        garbage_size = encoding_length - (offset - keep)
        #  self.garbage = data[offset: offset+garbage_size]
        offset += garbage_size

        self.qualifier_set.load(self.class_heap)
        self.properties.load(self.class_heap, self.nd_value_table)

        return self, offset

    def get_name(self):
        if self.class_name_ref == 0xffffffff:
            return 'None'

        heap = self.class_heap
        name, offset = EncodedString.from_data(heap, self.class_name_ref)

        names = [name]
        dlist = self.class_name_encoding
        offset, end = 0, len(dlist)
        while offset < end:
            super_class, offset = EncodedString.from_data(dlist, offset)
            names.append(super_class)

        return ':'.join(names)



from __future__ import annotations
from typing import TYPE_CHECKING, Callable
import struct
from .common import RpcCommon
from ..tools import pad4
from .result import RpcResult
from .auth_verifier_co import RpcAuthVerifierCo
from .cont_elem import RpcContElem
from ..uuid import bin_to_uuid_ver
from .baseresp import RpcBaseResp
from .const import RPC_C_AUTHN_LEVEL_PKT_PRIVACY
from .const import RPC_C_AUTHN_WINNT
from .const import RPC_C_AUTHN_GSS_NEGOTIATE
from ..exceptions import wbem_exception


if TYPE_CHECKING:
    from ..dcom import Dcom
    from ..protocol import Protocol

# /* common fields */
# u_int16 max_xmit_frag; /* 16:02 max transmit frag size */
# u_int16 max_recv_frag; /* 18:02 max receive frag size */
# u_int32 assoc_group_id; /* 20:04 returned assoc_group_id */
# port_any_t sec_addr; (length u16, char[]) /* 24:yy optional secondary address

# * for process incarnation; local port

# * part of address only */
# /* restore 4-octet alignment */
# u_int8 [size_is(align(4))] pad2;
# /* presentation context result list, including hints */
# p_result_list_t p_result_list; /* variable size */
# typedef struct {
# u_int8 n_results; /* count */

# u_int8 reserved; /* alignment pad, m.b.z. */

# u_int16 reserved2; /* alignment pad, m.b.z. */

# p_result_t [size_is(n_results)] p_results[];

# } p_result_list_t;
# /* optional authentication verifier */
# /* following fields present iff auth_length != 0 */
# auth_verifier_co_t auth_verifier; /* xx:yy */
# } rpcconn_bind_ack_hdr_t


class RpcResponse(RpcBaseResp):

    __slots__ = ('rpc_common', 'ctx_id', 'cancel_count', 'padding')

    RESPONSE_FMT = '<LHBB'
    RESPONSE_SIZE = struct.calcsize(RESPONSE_FMT)
    SIZE = RpcCommon.COMMON_SIZE + RESPONSE_SIZE

    ERR_FMT = '<L'

    def __init__(self, dcom: Dcom, rpc_common: RpcCommon, data: bytes):
        offset = RpcCommon.COMMON_SIZE
        (
            alloc_hint,
            self.ctx_id,
            self.cancel_count,
            self.padding
        ) = struct.unpack_from(
            self.RESPONSE_FMT,
            data,
            offset)

        self.rpc_common = rpc_common

    def get_message(self, proto: Protocol):
        messages = []
        for (resp, n) in self.get_pdu_data_list():
            auth_n = self.rpc_common.auth_length
            if auth_n:
                offset = n - (auth_n + RpcAuthVerifierCo.SIZE)
                auth = RpcAuthVerifierCo(resp, auth_n, offset)
                resp = resp[:offset]

                if auth.auth_level == RPC_C_AUTHN_LEVEL_PKT_PRIVACY:
                    if auth.auth_type == RPC_C_AUTHN_WINNT:
                        message, signature = proto._server_seal(
                            flags=proto._flags,
                            seq_num=proto._dcom._seq_num,
                            message_to_sign=resp,
                            message_to_encrypt=resp)
                    elif auth.auth_type == RPC_C_AUTHN_GSS_NEGOTIATE:
                        message = proto._server_seal(
                            cipher_text=resp,
                            auth_data=auth.auth_value)
                    else:
                        raise Exception(
                            f'Unsupported auth type: {auth.auth_type}')
                else:
                    message = resp

                if auth.auth_pad_length:
                    message = message[:-auth.auth_pad_length]
            else:
                # no authentication ??
                message = resp
            messages.append(message)

        if not message.endswith(b'\x00\x00\x00\x00'):
            # First, rpc status codes should also have an rpc fault package
            # and are already handled.
            # Second, WBEM_S "errors" are not relevant and can be distinguished
            # with the 0x80000000 bit, although, we do raise them
            errcode, = struct.unpack('<L', message[-4:])
            raise wbem_exception(errcode)

        return b''.join(messages)



import struct
import random
import hmac
import os
import hashlib
from datetime import datetime, timedelta, timezone
from .tools import aes_cts_encrypt, derive_key
from .asn1 import (
    asn1_len, asn1_seq, asn1_tag, asn1_int, asn1_gt, asn1_gs, asn1_ostr
)


def build_as_req(username: str, domain: str, pa_enc: bytes = b'') -> bytes:
    domain = domain.upper()

    # 0x01 0x01 0xff = Boolean True
    pac_req_inner = asn1_tag(0, b'\x01\x01\xff')
    pac_req = asn1_seq(pac_req_inner)

    # Type 128 (PA-PAC-REQUEST)
    padata_element_content = (
        asn1_tag(1, asn1_int(128)) +
        asn1_tag(2, asn1_ostr(pac_req))
    )
    padata_sequence_of = asn1_seq(pa_enc + asn1_seq(padata_element_content))

    # NT-PRINCIPAL = 1
    sname_components = (
        b'\x1b' + asn1_len(b'krbtgt') + b'krbtgt' +
        b'\x1b' + asn1_len(domain.encode()) + domain.encode()
    )
    sname = (
        asn1_tag(0, b'\x02\x01\x01') +
        asn1_tag(1,
                 b'\x30' +
                 struct.pack('B', len(sname_components)) +
                 sname_components)
    )

    # req-body: cname (user)
    cname_components = (
        b'\x1b' + asn1_len(username.encode()) + username.encode()
    )
    cname = (
        asn1_tag(0, asn1_int(1)) +  # NT-PRINCIPAL
        asn1_tag(1, asn1_seq(cname_components))
    )

    # KDC Body
    nonce = random.getrandbits(31)
    till = (
        datetime.now(timezone.utc) +
        timedelta(days=1)
    ).strftime("%Y%m%d%H%M%SZ").encode()

    # kdc-options: Forwardable, Proxiable, Renewable, Canonicalize
    kdc_options = b'\x03\x05\x00\x50\x80\x00\x00'
    req_body_fields = (
        asn1_tag(0, kdc_options) +
        asn1_tag(1, asn1_seq(cname)) +
        asn1_tag(2, b'\x1b' + asn1_len(domain.encode()) + domain.encode()) +
        asn1_tag(3, asn1_seq(sname)) +
        asn1_tag(5, asn1_gt(till)) +  # till
        asn1_tag(6, asn1_gt(till)) +  # rtime
        asn1_tag(7, asn1_int(nonce)) +
        # etype: aes256-cts-hmac-sha1-96 (18)
        asn1_tag(8, asn1_seq(asn1_int(18)))
    )
    req_body_sequence = asn1_seq(req_body_fields)

    # AS_REQ
    as_req_fields = (
        asn1_tag(1, b'\x02\x01\x05') +  # pvno
        asn1_tag(2, b'\x02\x01\x0a') +  # msg-type
        asn1_tag(3, padata_sequence_of) +  # padata (Tag 3)
        asn1_tag(4, req_body_sequence)  # req-body (Tag 4)
    )
    inner_seq = asn1_seq(as_req_fields)

    # APPLICATION TAG 10 (0x6a)
    final_as_req = b'\x6a' + asn1_len(inner_seq) + inner_seq
    return final_as_req


def build_full_as_req(username: str, domain: str, base_key: bytes):
    now = datetime.now(timezone.utc)
    ts_str = now.strftime("%Y%m%d%H%M%SZ").encode()
    original_plain_body = (
        asn1_tag(0, asn1_gt(ts_str)) +       # [0] pausec (GeneralizedTime)
        asn1_tag(1, asn1_int(0))             # [1] pusec (Microseconds, 0)
    )
    original_plain = asn1_seq(original_plain_body)

    confounder = os.urandom(16)
    full_plain = confounder + original_plain

    ke = derive_key(base_key, 1, 0xAA)
    ki = derive_key(base_key, 1, 0x55)
    signature = hmac.new(ki, full_plain, hashlib.sha1).digest()[:12]
    cipher_only = aes_cts_encrypt(ke, full_plain)

    final_payload = cipher_only + signature

    enc_data_content = (
        asn1_tag(0, asn1_int(18)) +         # etype 18 (AES-256)
        asn1_tag(2, asn1_ostr(final_payload))
    )

    enc_data = asn1_seq(enc_data_content)
    pa_ts_content = (
        asn1_tag(1, asn1_int(2)) +          # Type: PA-ENC-TIMESTAMP
        asn1_tag(2, asn1_ostr(enc_data))    # Value: Octet String
    )
    pa_ts = asn1_seq(pa_ts_content)

    pa_pac_val = asn1_seq(asn1_tag(0, b'\x01\x01\xff'))  # Boolean True
    pa_pac_content = (
        asn1_tag(1, asn1_int(128)) +
        asn1_tag(2, asn1_ostr(pa_pac_val))
    )
    pa_pac = asn1_seq(pa_pac_content)
    padata_field = asn1_tag(3, asn1_seq(pa_ts + pa_pac))

    # KDC-REQ-BODY
    domcaps = domain.upper()
    nonce = random.getrandbits(31)
    till = (now + timedelta(days=1)).strftime("%Y%m%d%H%M%SZ").encode()

    cname_content = (
        asn1_tag(0, asn1_int(1)) +
        asn1_tag(1, asn1_seq(asn1_gs(username.encode())))
    )
    sname_content = (
        asn1_tag(0, asn1_int(1)) +
        asn1_tag(1, asn1_seq(asn1_gs(b"krbtgt") + asn1_gs(domcaps.encode())))
    )

    req_body_content = (
        asn1_tag(0, b'\x03\x05\x00\x50\x80\x00\x00') +  # kdc-options
        asn1_tag(1, asn1_seq(cname_content)) +
        asn1_tag(2, asn1_gs(domcaps.encode())) +
        asn1_tag(3, asn1_seq(sname_content)) +
        asn1_tag(5, asn1_gt(till)) +
        asn1_tag(6, asn1_gt(till)) +
        asn1_tag(7, asn1_int(nonce)) +
        asn1_tag(8, asn1_seq(asn1_int(18)))             # etype list
    )

    as_req_content = (
        asn1_tag(1, asn1_int(5)) +                      # pvno
        asn1_tag(2, asn1_int(10)) +                     # msg-type (AS-REQ)
        padata_field +
        asn1_tag(4, asn1_seq(req_body_content))         # req-body
    )

    inner_sequence = asn1_seq(as_req_content)
    final_as_req = b'\x6a' + asn1_len(inner_sequence) + inner_sequence
    return final_as_req



from __future__ import annotations
from typing import TYPE_CHECKING
import struct
from .common import RpcCommon
from ..tools import pad4
from .result import RpcResult
from .auth_verifier_co import RpcAuthVerifierCo
from .cont_elem import RpcContElem
from ..uuid import bin_to_uuid_ver
from .baseresp import RpcBaseResp
from ..exceptions import rpc_exception
from ..logger import logger


if TYPE_CHECKING:
    from ..dcom import Dcom


class RpcFault(RpcBaseResp):
    __slots__ = ('status', 'rpc_common')

    FAULT_FMT = '<III'
    FAULT_FMT_SIZE = struct.calcsize(FAULT_FMT)

    def __init__(self, dcom: Dcom, rpc_common: RpcCommon, data: bytes):
        self.rpc_common = rpc_common
        body_data = data[RpcCommon.COMMON_SIZE:]
        self.status = 0x00000721   # RPC_S_SEC_PKG_ERROR (most likely)
        if len(body_data) >= self.FAULT_FMT_SIZE:
            _, _, self.status = struct.unpack(self.FAULT_FMT, body_data[:12])
        elif len(body_data) >= 4:
            alloc_hint = struct.unpack('<I', body_data[:4])[0]
            logger.debug(f"RPC Fault received with alloc_hint: {alloc_hint}")

    def throw(self):
        raise rpc_exception(self.status)



import struct
from .version import NTLMVersion
from .const import (
    NTLMSSP_NEGOTIATE_128,
    NTLMSSP_NEGOTIATE_56,
    NTLMSSP_NEGOTIATE_ALWAYS_SIGN,
    NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY,
    NTLMSSP_NEGOTIATE_KEY_EXCH,
    NTLMSSP_NEGOTIATE_NTLM,
    NTLMSSP_NEGOTIATE_OEM_DOMAIN_SUPPLIED,
    NTLMSSP_NEGOTIATE_OEM_WORKSTATION_SUPPLIED,
    NTLMSSP_NEGOTIATE_SEAL,
    NTLMSSP_NEGOTIATE_SIGN,
    NTLMSSP_NEGOTIATE_TARGET_INFO,
    NTLMSSP_NEGOTIATE_UNICODE,
    NTLMSSP_NEGOTIATE_VERSION,
    NTLMSSP_REQUEST_TARGET,
)

# [MS-NLMP.pdf]  2.2.1.1 NEGOTIATE_MESSAGE
# {
#   fixed:                      NTLMSSP + \0
#   message_type:               <L MUST be 1
#   negotiate_flags:            <L flags
#   domain_name_len:            <H len domain name
#   domain_name_max_len:        <H Equal to  domain_name_len
#   domain_name_offset:         <L Offset to buffer in payload, from begin
#   workstation_name_len:       <H len workstation name
#   workstation_name_max_len:   <H Equal to workstation_name_len
#   workstation_name_offset:    <L Offset to buffer in payload, from begin
#   version:                    Optional 8 bytes (Version structure)
#   payload {
#       domain_name,
#       workstation_name,
#   }
# }


class NTLMAuthNegotiate:
    """Message from client to server."""

    SIGNATURE = b'NTLMSSP\x00'
    SIGNATURE_LEN = len(SIGNATURE)
    MESSAGE_TYPE = 1
    NEGOTIATE_FMT = '<LLHHLHHL'
    NEGOTIATE_SIZE = SIGNATURE_LEN + struct.calcsize(NEGOTIATE_FMT)

    def __init__(self):
        self._negotiate_flags = (
            # Sign in required flags
            NTLMSSP_NEGOTIATE_KEY_EXCH |
            NTLMSSP_NEGOTIATE_SIGN |
            NTLMSSP_NEGOTIATE_ALWAYS_SIGN |
            NTLMSSP_NEGOTIATE_SEAL |
            # NTLM v2 flags
            NTLMSSP_NEGOTIATE_TARGET_INFO |
            # Always
            NTLMSSP_NEGOTIATE_NTLM |
            NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY |
            NTLMSSP_NEGOTIATE_UNICODE |
            NTLMSSP_REQUEST_TARGET |
            NTLMSSP_NEGOTIATE_128 |
            NTLMSSP_NEGOTIATE_56)
        self._version = b''
        self._version_len = 0
        self._domain_name = b''
        self._domain_name_len = 0
        self._workstation_name = b''
        self._workstation_name_len = 0

    def get_negotiate_flags(self):
        return self._negotiate_flags

    def set_version(self, version: NTLMVersion):
        """Version is packed as 8 bytes.
        So:
          len(self._version) ==  NTLMVersion.VERSION_SIZE == 8 bytes
        """
        self._version = version.get_data()
        self._version_len = NTLMVersion.VERSION_SIZE
        self._negotiate_flags |= NTLMSSP_NEGOTIATE_VERSION

    def set_domain_name(self, domain_name):
        self._domain_name = domain_name.encode('utf-16le')
        self._domain_name_len = len(self._domain_name)
        self._negotiate_flags |= NTLMSSP_NEGOTIATE_OEM_DOMAIN_SUPPLIED

    def set_workstation_name(self, workstation_name):
        self._workstation_name = workstation_name.encode('utf-16le')
        self._workstation_name_len = len(self._workstation_name)
        self._negotiate_flags |= NTLMSSP_NEGOTIATE_OEM_WORKSTATION_SUPPLIED

    def get_data(self) -> bytes:
        domain_name_offset = self.NEGOTIATE_SIZE + self._version_len
        workstation_name_offset = domain_name_offset + self._domain_name_len

        data = self.SIGNATURE + struct.pack(
            self.NEGOTIATE_FMT,
            self.MESSAGE_TYPE,
            self._negotiate_flags,
            self._domain_name_len,
            self._domain_name_len,  # max must be equal to len
            domain_name_offset,
            self._workstation_name_len,
            self._workstation_name_len,  # max must be equal to len
            workstation_name_offset,
        ) + self._version + self._domain_name + self._workstation_name
        return data
